#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/atomic.h>

#include "adc.h"

// Der Reload von T1 triggert den ADC!

#define ADC_CHANNEL_MASK (_adc_ch_next_power_of_2-1)
#define ADC_DDR DDRC
#define ADC_PORT PORTC

// Der Vorteiler muss also so eingestellt werden, dass CPU-Taktfrequenz dividiert durch den Teilungsfaktor einen Wert im Bereich (50-200)kHz ergibt. 
#if (F_CPU == 20000000UL)
// Teiler Min= 20000000/200000 = 100
// Teiler Max= 20000000/50000= 400
 #define ADC_DIV 128 
#else
 #error define DADC Divisor for F_CPU
#endif

TAdcScaler AdcValues[ADC_CHANNELS];

void TAdcScaler::SetScaler(uint8_t mean, uint16_t _max)
{
	MeanMax = mean;
	Dividend = _max;
	Divisor = 1023U;
	Divisor *= mean;
	MeanSum = 0;
	MeanCount = MeanMax;
}

TMeasResult TAdcScaler::GetMean(bool round)
{
	TMeasResult res;
	if (MeanCount == 0)
	{
		uint32_t x;
		ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
		{
			x = MeanSum;
			MeanSum = 0;
			MeanCount = MeanMax;
		}
		x *= Dividend;
		if (round) x += Divisor / 2;
		x /= Divisor;
		res.Value = (uint16_t)x;
		res.Valid = 1;
		return res;
	}
	res.Value = 0;
	res.Valid = 0;
	return res;
}

ISR(ADC_vect)
{
	uint16_t value = ADCW;

	// Channel = (Channel+1) % ADC_CHANNELS
	uint8_t m = ADMUX;
	uint8_t i = m & ADC_CHANNEL_MASK;
	// Zwischen den Channel-Bits und den Konfig-Bits gibt es ein Bit was immer 0 gelesen wird.
	// Deshalb ist das increment der (unteren) Channel-Bits gefahrlos mglich!
	m++;
	// Achtung m % ADC_CHANNELS ist nur sinnvoll, wenn ADC_CHANNELS Power of 2 ist!!!
	m = (m & 0xE0) | (m % ADC_CHANNELS);
	ADMUX = m;

	// AD-Wert verarbeiten (Aufsummierung MeanRegister)
	// Das hier ist eine friend-Funktion von TAdcScaler!
	TAdcScaler& adc = AdcValues[i];
	uint8_t c = adc.MeanCount;
	if (c)
	{
		adc.MeanCount = c - 1;
		adc.MeanSum += value;
	}
}

inline void SetChannel(uint8_t ch)
{
	ADMUX = (uint8_t)(ADMUX & 0xf0) | ch;
}

void TAdcScaler::Init()
{
	// Analog Comparator off
	ACSR &= ~(1 << ACIE);
	ACSR |= (1 << ACD);

	// Port als Eingang, hochohmig
	ADC_DDR &= ~ADC_PIN_MASK;
	ADC_PORT &= ~ADC_PIN_MASK;
	DIDR0 |= ADC_PIN_MASK;			// Disable digital Pin

	ADMUX = ADC_REF_SOURCE;
	ADCSRB = 0;						// Free running mode

#if (ADC_DIV==2)
	ADCSRA = (1 << ADPS0);
#elif (ADC_DIV==4)
	ADCSRA = (1 << ADPS1);
#elif (ADC_DIV==8)
	ADCSRA = (1 << ADPS1) | (1 << ADPS0);
#elif (ADC_DIV==16)
	ADCSRA = (1 << ADPS2);
#elif (ADC_DIV==32)
	ADCSRA = (1 << ADPS2) | (1 << ADPS0);
#elif (ADC_DIV==64)
	ADCSRA = (1 << ADPS2) | (1 << ADPS1);
#elif (ADC_DIV==128)
	ADCSRA = (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0);
#else
#error define DADC Divisor 
#endif	

	ADCSRA |= (1 << ADEN);						// ADC aktivieren

	/* nach Aktivieren des ADC wird ein "Dummy-Readout" empfohlen, man liest
	   also einen Wert und verwirft diesen, um den ADC "warmlaufen zu lassen" */
	trigger_adc();							// eine ADC-Wandlung 
	while (ADCSRA & (1 << ADSC)) {}			// auf Abschluss der Konvertierung warten
	/* ADCW muss einmal gelesen werden, sonst wird Ergebnis der nchsten
	   Wandlung nicht bernommen. */
	uint16_t dummy = ADCW;
	ADCSRA |= (1 << ADIF);						// Int. Req. Bit lschen
	SetChannel(0);
	if (dummy);
}

void TAdcScaler::Start()
{
	ADCSRA |= (1 << ADIE);						// Interrupt aktivieren 

#ifndef SIMULATION	
	uint8_t m = 0;
	while (m != (1 << ADC_CHANNELS) - 1)				// warten, bis alle Kanle einmal gelesen wurden
	{
		for (uint8_t i = 0; i < ADC_CHANNELS; i++)
		{
			if (AdcValues[i].GetMean(false).Valid)
				m |= (1 << i);
		}
	}
#endif
}


