#include <avr/io.h>
#include <avr/interrupt.h>

#include <stdlib.h>

#include "dc.h"
#include "ac.h"
#include "adc.h"
#include "strfmt.h" 
#include "status.h"

// Zwischenkreisregelung
// Alle Spannungen in 1/10 Volt!!!
// Benutzt Timer 0 im Fast PWM Modus mit CTC

#define F_TRAFO 50000UL					/* 50 kHz */
#define F_RELOAD 100000UL				/* wird in Hardware / 2 geteilt -> doppelte Frequenz ausgeben */
#define PWM_RELOAD (F_CPU/F_RELOAD)	
#define PWM_MIN 0						/* der Minimalwert, der in OCR0B geladen werden kann */
#define PWM_MAX (PWM_RELOAD-1)			/* der Maximalwert, der in OCR0B geladen werden kann */
#define TRAFO_MAX_VOLTAGE 3500			/* der Trafo bringt max. 350 V */

// Reglerkonstanten
#define Q_SCALER 128
const float Kp = 20;//10;//6;//5;					// Proportionalbeiwert
const float Ki = 40;//30;//25;//15;//8;//4;//3;					// Integralbeiwert
const float Kd = 0.05;					// Differentialbeiwert

static int32_t q0, q1, q2;				// Faktoren aus dem Stellungsalgorithmus
static int16_t yalt;
int16_t CurrentDcVoltage;

static void InitTimer0()
{
	// Timer 0 macht F_TRAFO*2 kHz PWM Frequenz fr die Primransteuerung des Trafos
	// Die PWM wird in 2 phasenverschobene Signale gewandelt (Teilung / 2) 
	// -> effektive Ansteuerung des Trafos mit F_TRAFO kHz
	// kein Prescaler, PWM an OC0B, 
	// Fast PWM TOP=OCRA = WGM 7
	TCCR0A = (1 << COM0B1) | (1 << WGM01) | (1 << WGM00);
	TCCR0B = (1 << WGM02);
	OCR0A = PWM_RELOAD;
	OCR0B = 0;
	DC_CTRL_DDR |= (1 << DC_PWM_PIN);			// OC0B PIN als output -> PWM
}

void InitDc()
{
	// Steuer-Ausgnge als Output, Pegel Low
	DisableDc();
	DC_CTRL_DDR |= (1 << DC_ENABLE_PIN);

	InitTimer0();

	yalt = 0;
	SetDcPwm(yalt);

	// Regler-Faktoren berechnen
	float Ta = ADC_SAMPLE_MICROS(DC_ADC_MEAN) / 1e6;
	float q = Kp + Ki*Ta + Kd / Ta;
	q0 = (int32_t)(q*Q_SCALER + 0.5);

	q = -(Kp + 2 * Kd / Ta);
	q1 = (int32_t)(q*Q_SCALER - 0.5);

	q = Kd / Ta;
	q2 = (int32_t)(q*Q_SCALER + 0.5);
}

void StartDc()
{
	TCCR0B |= (1 << CS00);				// Timer starten
}

void SetDcPwm(uint16_t uout)
{
	// der Trafo soll TRAFO_MAX_VOLTAGE bei 100% PWM abgeben
	// 100% PWM sind PWM_RELOAD

	//       uout * PWM_RELOAD
	// pwm= ____________________
	//       TRAFO_MAX_VOLTAGE

	// um mit 16 bit auszukommen wird daraus:

	//       uout * PWM_RELOAD/100
	// pwm= ________________________
	//       TRAFO_MAX_VOLTAGE/100

	uint16_t pwm = uout;
	pwm *= (PWM_RELOAD / 100);
	pwm /= (TRAFO_MAX_VOLTAGE / 100);

	if (pwm > PWM_MAX)		OCR0B = (uint8_t)PWM_MAX;
	else if (pwm < PWM_MIN)	OCR0B = (uint8_t)PWM_MIN;
	else					OCR0B = (uint8_t)pwm;
}

#define LIMIT_CONTROL	10000
static void DoRegulation(int16_t voltage)
{
	static int16_t ealt = 0;
	static int16_t ealt2 = 0;

	int16_t e = TARGET_VOLTAGE - voltage;					// Vergleich

	//Reglergleichung 

	int32_t tmp = q0*e + q1*ealt + q2*ealt2;
	tmp /= Q_SCALER;										// die q-Faktoren sind mit Q_SCALER skaliert, zurckrechnen!
	tmp += yalt;

	// Begrenzung fr int16
	if (tmp < -LIMIT_CONTROL) tmp = -LIMIT_CONTROL;
	else if (tmp > LIMIT_CONTROL) tmp = LIMIT_CONTROL;
	int16_t y = (int16_t)tmp;

	ealt2 = ealt;
	ealt = e;
	yalt = y;

	// Ausgabe
	// Begrenzung Output (PWM)
	if (y < 0) y = 0;
	else if (y > TRAFO_MAX_VOLTAGE) y = TRAFO_MAX_VOLTAGE;
	SetDcPwm(y);
}


// Aufruf aller ADC_SAMPLE_MICROS(DC_ADC_MEAN) s = 2 ms
void HandleDc(int16_t voltage)
{
	static uint8_t last_state = -1;
	CurrentDcVoltage = voltage;

	uint8_t s = MachineState.DoTimer(voltage);

	if (mstReg == s)
	{
		if (mstReg != last_state)
			yalt = voltage;				// Preset Regler
		DoRegulation(voltage);
	}
	last_state = s;
}


