#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/atomic.h>
#include <stdlib.h>
#include <avr/wdt.h>

extern "C"	{
#include "lcd.h"
#include "led.h"
#include "strfmt.h" 
#include "keys.h"
	extern volatile uint16_t timer_100;		// wird von Timer1 Reload ISR in ac.c getrieben
};

char LcdBuff[16 + 1];

#include "ac.h"
#include "dc.h"
#include "eedata.h"
#include "Adc.h"
#include "status.h"
#include "MeasValue.h"
#include "Menue.h"
#include "Iir.h"
#include "Utils.h"

inline uint16_t ReadTimer100()
{
	ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
	{
		return timer_100;
	}
	return 0;	// avoid warning
}

inline uint16_t ElapsedTimer100(uint16_t since)
{
	uint16_t diff = ReadTimer100() - since;	// both operands are uint16_t -> overflow is handled automatically
	return diff;
}

// delay of 10 ms
void Delay10(uint16_t ms10)
{
	uint16_t t = ReadTimer100();
	while (ElapsedTimer100(t) < ms10) 
	{
		wdt_reset();
	}
}


// http://www.nongnu.org/avr-libc/user-manual/FAQ.html#faq_rom_array
static const char ShowTx0[] PROGMEM = "UA: ";
static const char ShowTx1[] PROGMEM = "IE: ";
static const char ShowTx2[] PROGMEM = "UE: ";
static const char ShowTx3[] PROGMEM = "IA: ";
static const char ShowTx4[] PROGMEM = "PE: ";
static const char ShowTx5[] PROGMEM = "PA: ";
static const char ShowTx6[] PROGMEM = "Wirk.Gr.: ";

static const char ErrTx0[] PROGMEM = "\xf5""berspannung ZK!";
static const char ErrTx1[] PROGMEM = "Eingangsstrom!";
static const char ErrTx2[] PROGMEM = "Eing. spannung!";
static const char ErrTx3[] PROGMEM = "Ausgangsstrom!";
static const char ErrTx4[] PROGMEM = "Wirkungsgrad!";

static const char UnitTx_A[] PROGMEM = "A";
static const char UnitTx_V[] PROGMEM = "V";
static const char UnitTx_W[] PROGMEM = "W";
static const char UnitTx_Prc[] PROGMEM = "%";

#define INVALID_MONITOR 0xff
class TEEMonitorValue : public TMonitorValue {
public:
	TEEMonitorValue(uint8_t digits, uint8_t divide, uint8_t hysteresis,
		PGM_P descrTx, PGM_P unitTx, PGM_P errorTx)
		: TMonitorValue(digits, divide, hysteresis, descrTx, unitTx, errorTx) {}

	void SetFromEEprom(uint8_t eeLimit, uint8_t multiply, Directions direction,
		uint16_t timeTicks, uint8_t strictPercent = 0, uint16_t strictTicks = 0)
	{
		if (eeLimit == INVALID_MONITOR)
			Disable();
		else
			SetMonitor(eeLimit*multiply, direction, timeTicks, strictPercent, strictTicks);
	}
};

//								digits	divide	hysteresis	Descr.		Unit		Error
static TMonitorValue UOutMoni(	0,		1,		2,			ShowTx0,	UnitTx_V,	ErrTx0);
static TEEMonitorValue IInMoni(	1,		0,		2,			ShowTx1,	UnitTx_A,	ErrTx1);
static TEEMonitorValue UInMoni(	1,		0,		2,			ShowTx2,	UnitTx_V,	ErrTx2);
static TEEMonitorValue IOutMoni(2,		0,		2,			ShowTx3,	UnitTx_A,	ErrTx3);
static TMeasValue PInValue(		0,		0,		2,			ShowTx4,	UnitTx_W);
static TMeasValue POutValue(	0,		1,		2,			ShowTx5,	UnitTx_W);
static TEEMonitorValue EceMoni(	0,		1,		2,			ShowTx6,	UnitTx_Prc, ErrTx4);

static IirFilter UOutFilter(32);
static IirFilter IInFilter(32);
static IirFilter UInFilter(32);
static IirFilter IOutFilter(32);

enum eMeasFlags	{
	mfUIn,
	mfUOut,
	mfIIn,
	mfIOut,
	mfPIn,
	mfPOut,
	mfEce,	// Energy conversion efficiency = Wirkungsgrad
	_mfLast,
};

static TMeasValue* const pMeasureValues[_mfLast] PROGMEM = {
	&UInMoni,
	&UOutMoni,
	&IInMoni,
	&IOutMoni,
	&PInValue,
	&POutValue,
	&EceMoni,
};

static void InitIO()
{
	// Fr Ausgangspins, die mit Anfangswert "high" initialisiert werden sollen: 
	// - zuerst die Bits im PORTx-Register setzen
	// - anschlieend die Datenrichtung auf Ausgang stellen

	PORTB = ~(1 << AC_ENABLE_PIN);	// unused pull ups in general on
	PORTC = ~ADC_PIN_MASK;			// unused pull ups in general on
	PORTD = ~(1 << DC_ENABLE_PIN);	// unused pull ups in general on
	DDRB = 0;						// all inputs
	DDRC = 0;						// all inputs
	DDRD = 0;						// all inputs
}

void SetMonitoring()
{
	IInMoni.SetFromEEprom(eeData.max_inp_current,	10, TMonitorValue::Plus,	ADC_SAMPLE_NUM(50, 5000), 10, ADC_SAMPLE_NUM(50, 2000));
	IOutMoni.SetFromEEprom(eeData.max_outp_current,	10, TMonitorValue::Plus,	ADC_SAMPLE_NUM(50, 5000), 10, ADC_SAMPLE_NUM(50, 2000));
	UInMoni.SetFromEEprom(eeData.min_inp_voltage,	1, TMonitorValue::Minus,	ADC_SAMPLE_NUM(50, 5000), 10, ADC_SAMPLE_NUM(50, 2000));
	EceMoni.SetFromEEprom(eeData.min_ece,			10, TMonitorValue::Minus,	ADC_SAMPLE_NUM(50, 10000));
}

//////////////// Men //////////////
static void MenuFmtDecimal(const TMenuContext& ctx, char* buff, uint8_t ndp)
{
	if (ctx.HoldsInvalid()) strcpy_P(buff, PSTR("---"));
	else
	{
		itoa(ctx.GetValue(), buff, 10);
		if (ndp) set_dec_point(buff, ndp);
	}
}


static void MenuFmtDecimal1(const TMenuContext& ctx, char* buff)
{
	MenuFmtDecimal(ctx, buff, 0);
}

static void MenuFmtDecimal10(const TMenuContext& ctx, char* buff)
{
	MenuFmtDecimal(ctx, buff, 1);
}

static void MenuFmtBool(const TMenuContext& ctx, char* buff)
{
	if (ctx.GetValue()) strcpy_P(buff, PSTR("Ein"));
	else strcpy_P(buff, PSTR("Aus"));
}

static const char MnDescr_0[] PROGMEM = "Eing.-Strom Max.";
static const char MnDescr_1[] PROGMEM = "Ausg.-Strom Max.";
static const char MnDescr_2[] PROGMEM = "Eing.-Spg. Min.";
static const char MnDescr_3[] PROGMEM = "Wirk.-Grad Min.";
static const char MnDescr_4[] PROGMEM = "Autostart";

#define MENU_NUM 5
#define MENU_OFF 0xff

static const TMenuItem MenuItems[MENU_NUM] PROGMEM = {
	//	MinVal	MaxVal	InvVal				Flags		Step	EEAddr						Format				DescrTx		UnitTx
	{	2,		55,		INVALID_MONITOR,	mnfInvDown, 1,		&eeData.max_inp_current,	MenuFmtDecimal1,	MnDescr_0,	UnitTx_A, },
	{	2,		30,		INVALID_MONITOR,	mnfInvDown, 2,		&eeData.max_outp_current,	MenuFmtDecimal10,	MnDescr_1,	UnitTx_A, },
	{	80,		120,	INVALID_MONITOR,	mnfInvDown, 1,		&eeData.min_inp_voltage,	MenuFmtDecimal10,	MnDescr_2,	UnitTx_V, },
	{	50,		100,	INVALID_MONITOR,	mnfInvUp,	1,		&eeData.min_ece,			MenuFmtDecimal1,	MnDescr_3,	UnitTx_Prc, },
	{	0,		1,		INVALID_MONITOR,	mnfWrap,	1,		&eeData.auto_on,			MenuFmtBool,		MnDescr_4, },
};


class MyMenuContext : TMenuContext	{
private:
	bool		Changes;
	uint16_t	Time;

public:
	MyMenuContext() { Changes = false; Time = ReadTimer100(); }
	uint8_t HandleMenue(uint8_t idx, uint8_t keys);
};

uint8_t MyMenuContext::HandleMenue(uint8_t idx, uint8_t keys)
{
	if (idx == MENU_OFF)
	{
		// start
		Changes = false;
		Time = ReadTimer100();
		set_key_repeat((1 << KEY_ON) | (1 << KEY_OFF));
		keys |= (1 << KEY_MENU);
	}
	if (keys & (1 << KEY_MENU))
	{
		if (idx < MENU_NUM)
		{
			if (EndEdit()) Changes = true;
		}
		idx++;
		if (idx == MENU_NUM)
		{
			// fertig
			if (Changes)
			{
				write_ee(&eeData);
				SetMonitoring();
			}
			set_key_repeat(0);
			return MENU_OFF;
		}
		StartEdit(MenuItems[idx], 0, 0);
	}
	else if (keys & (1 << KEY_ON)) DoEdit(TMenuContext::Up);
	else if (keys & (1 << KEY_OFF)) DoEdit(TMenuContext::Down);
	else
	{
		// keine Aktion
		if (ElapsedTimer100(Time) > 3000) idx = MENU_OFF;	// zu lange keine Aktion mehr -> zurck
		return idx;
	}
	// Anzeige aktualisieren
	Print(0, 1);
	Time = ReadTimer100();
	return idx;
}
//////////////// Men //////////////////


bool CheckMonitor(TMonitorValue& m, uint16_t value, bool doCheck)
{
	if (!m.SetValue(value) && doCheck)
	{
		MachineState.SetError(mstError);
		m.PrintLastError1(0, 0);
		m.PrintLastError2(0, 1);
		return false;
	}
	return true;
}

enum ePrintState	{
	psNoValue,
	psDo,
	psDid,
};

int __attribute__((OS_main))
main(void)
{
	// Allerdings werden die Prescaler Bits gelscht, weshalb der Watchdog nach 16ms einen Reset
	// auslst, selbst wenn man den Watchdog z.B. mit 2s gestartet hat !!!
	
	wdt_enable(WDTO_250MS);
	
	InitIO();
	init_ee(&eeData);
	init_ac();
	InitDc();
	init_led();
	init_keys();
	TAdcScaler::Init();

	AdcValues[adc_ch_350V].SetScaler(DC_ADC_MEAN, 4044 - 80);	// ADC max entspricht theoretisch 404,4333V
	AdcValues[adc_ch_12V].SetScaler(50, 150/* 145*/);			// ADC max entspricht 14,972V (14,544)
	AdcValues[adc_ch_Iout].SetScaler(50, 240 - 30);				// ADC max entspricht 2,4A
	AdcValues[adc_ch_Iin].SetScaler(50, 720 - 70);				// ADC max entspricht theoretisch 72A

	UOutMoni.SetMonitor(MAX_INPUT_VOLTAGE, TMonitorValue::Plus, ADC_SAMPLE_NUM(DC_ADC_MEAN, 500), 10, ADC_SAMPLE_NUM(DC_ADC_MEAN, 250));
	SetMonitoring();

	sei();

	start_ac();							// ac macht auch den Timer fr delay und den ADC, zuerst starten!
#ifndef SIMULATION
	Delay10(100);
	lcd_init();
#endif
	TAdcScaler::Start();
	StartDc();
	lcd_clear();

	uint16_t dispTime = ReadTimer100();
	uint8_t dispIdx = 0;
	uint8_t menuIdx = MENU_OFF;
	uint8_t lastState = -1;
	uint8_t state = MachineState.GetState();
	ePrintState printStatex = psDo;
	MyMenuContext menu;

	while (1)
	{
#ifdef SIMULATION
		uint8_t keys = 0;
#else 
		uint8_t keys = get_key_press((1 << KEY_ON) | (1 << KEY_OFF) | (1 << KEY_MENU));
#endif
		wdt_reset();
		
		if (menuIdx != MENU_OFF ||
			((mstReady == state) && (keys & (1 << KEY_MENU))))
		{
			keys |= get_key_rpt((1 << KEY_ON) | (1 << KEY_OFF));
			menuIdx = menu.HandleMenue(menuIdx, keys);
			if (menuIdx == MENU_OFF) lastState = -1;
			keys = 0;
		}
		if (menuIdx == MENU_OFF)
		{
			if (keys & (1 << KEY_ON)) MachineState.Start();
			if (keys & (1 << KEY_OFF)) MachineState.Stop();
		}
		state = MachineState.GetState();
		if (state != lastState)
		{
			lastState = state;
			dispTime = ReadTimer100();
			if (menuIdx == MENU_OFF && !TStateMachine::IsError(state))
			{
				// Status gendert -> Betriebszustand ausgeben
				lcd_clear();
				PGM_P msg;
				switch (state)
				{
				case mstReady:			msg = PSTR("Bereit"); break;
				case mstStarting:		msg = PSTR("Starte..."); break;
				case mstStopping:		msg = PSTR("Stoppe...");
					// Messwerte, die nur bei Running Sinn machen disablen
					UOutMoni.ClearValue();
					PInValue.ClearValue();
					POutValue.ClearValue();
					EceMoni.ClearValue();
					break;
				case mstReg:			msg = PSTR("Betrieb");
					// Monitore resetten
					UOutMoni.Reset();
					IInMoni.Reset();
					UInMoni.Reset();
					IOutMoni.Reset();
					EceMoni.Reset();
					break;
				default:				msg = NULL;
				}
				if (msg) lcd_string_P(msg);
			}
		}

		// ZK-Spannung
		TMeasResult res = AdcValues[adc_ch_350V].GetMean(true);
		if (res.Valid)
		{
			HandleDc(res.Value);
			res.Value = UOutFilter.Execute(res.Value);
			if (!CheckMonitor(UOutMoni, res.Value, mstReg == state))
				continue;

			// zeitkritische DC Regelung wurde gerade ausgefhrt
			// rechenintensive Operationen sollten jetzt ausgefhrt werden

			if (mstReg == state)
			{
				// Leistungen und Wirkungsgrad nur wenn Running
				if (UInMoni.HasCalcValue() && IInMoni.HasCalcValue())
				{
					// Leistung Eingang
					uint32_t p = UInMoni.GetCalcValue();
					p *= IInMoni.GetCalcValue();
					p /= 100;	// Watt
					PInValue.SetValue((uint16_t)p);
				}

				if (UOutMoni.HasCalcValue() && IOutMoni.HasCalcValue())
				{
					// Leistung Ausgang
					uint32_t p = UOutMoni.GetCalcValue();
					p *= IOutMoni.GetCalcValue();
					p /= 100;	// 1/10Watt
					POutValue.SetValue((uint16_t)p);
				}

				if (PInValue.HasCalcValue() && POutValue.HasCalcValue())
				{
					// Wirkungsgrad
					uint32_t n = POutValue.GetCalcValue();
					n *= 100;
					uint16_t pi = PInValue.GetCalcValue();
					if (pi) n /= pi;
					else n = 0;
					if (!CheckMonitor(EceMoni, (uint16_t)n, mstReg == state))
						continue;
				}
			}

			// LCD Anzeige aktueller Wert
			if (menuIdx == MENU_OFF &&  !TStateMachine::IsError(state))
			{

				if ((printStatex == psNoValue) ||			// kein Wert vorhanden
					(ElapsedTimer100(dispTime) > 300) ||	// oder Timer
					(keys & (1 << KEY_MENU)))				// oder Menu-Taste
				{
					// schaltet auf nchsten Wert
					if (++dispIdx >= _mfLast) dispIdx = 0;
					dispTime = ReadTimer100();
					printStatex = psDo;
				}
				TMeasValue* pMeas = pgm_read_ptr(pMeasureValues[dispIdx]);
				// Wenn fr aktuellen Messwert schon eine Ausgabe gemacht wurde, 
				// Messwert bis zum Timeout beibehalten
				// sonst wenn kein Wert vorhanden, weiterschalten
				if (pMeas->Print(0, 1)) printStatex = psDid;
				else if (printStatex == psDo) printStatex = psNoValue;
			}
		}

		// Eingangstrom
		res = AdcValues[adc_ch_Iin].GetMean(true);
		if (res.Valid)
		{
			res.Value = IInFilter.Execute(res.Value);
			if (!CheckMonitor(IInMoni, res.Value, mstReg == state))
				continue;
		}

		// Eingangsspannung
		res = AdcValues[adc_ch_12V].GetMean(true);
		if (res.Valid)
		{
			res.Value = UInFilter.Execute(res.Value);
			if (!CheckMonitor(UInMoni, res.Value, mstReg == state))
				continue;
		}

		// Ausgangsstrom
		res = AdcValues[adc_ch_Iout].GetMean(true);
		if (res.Valid && (mstReg == state))	// nur wenn Running
		{
			res.Value = IOutFilter.Execute(res.Value);
			if (!CheckMonitor(IOutMoni, res.Value, true))
				continue;
		}
	}
}