#include <avr/io.h>

#include <stdlib.h>

#include "Utils.h"
#include "MeasValue.h"

extern "C"	{
#include "strfmt.h" 
#include "lcd.h"	
}

extern char LcdBuff[16 + 1];

/////////////////// TMeasValue ////////////////////////
TMeasValue::TMeasValue(uint8_t digits, uint8_t divide, uint8_t hysteresis,
	PGM_P _descrTx, PGM_P _unitTx)
{
	Info.Flags = 0;
	Info.Digits = digits;
	Info.Divide = divide;
	DescrTx = _descrTx;
	UnitTx = _unitTx;
	Value = 0;
#ifdef WITH_HYSTERESIS
	HystSpan = hysteresis;
	HystLow = HystUp = 0;
#endif
}

uint8_t TMeasValue::Format(uint16_t value, char* buff) const
{
	uint8_t i = Info.Divide;
	if (i > 0)
	{
		uint8_t div = 1;
		while (i--) div *= 10;
		value = (value + div / 2) / div;
	}
	itoa(value, buff, 10);
	return set_dec_point(buff, Info.Digits);
}

bool TMeasValue::SetValue(uint16_t _value)
{
#ifdef WITH_HYSTERESIS
	if (_value > HystUp)
	{
		HystUp = _value;
		if (_value < HystSpan) HystLow = 0;
		else HystLow = _value - HystSpan;
	}
	else if (_value < HystLow)
	{
		HystLow = _value;
		HystUp = _value + HystSpan;
	}
#endif
	Value = _value;
	Info.Flags |= flAll;
	return true;
}

bool TMeasValue::Print(uint8_t x, uint8_t y)
{
	if (Info.Flags & flMeas)
	{
		Info.Flags &= ~flMeas;
		lcd_setcursor(x, y);
		x += lcd_string_P(DescrTx);
#ifdef WITH_HYSTERESIS
		uint16_t val = (uint16_t)(((uint32_t)HystLow + (uint32_t)HystUp) / 2);
#else
		uint16_t val = Value;
#endif
		Format(val, LcdBuff);
		x += lcd_string(LcdBuff);
		if (UnitTx) x += lcd_string_P(UnitTx);
		while (x++ < 16) lcd_data(' ');
		return true;
	}
	return false;
}

uint16_t TMeasValue::GetCalcValue()
{
	Info.Flags &= ~flCalc;
	return Value;
}



/////////////////// TMonitorValue ////////////////////////
TMonitorValue::TMonitorValue(uint8_t digits, uint8_t divide, uint8_t hysteresis,
	PGM_P descrTx, PGM_P unitTx, PGM_P _errorTx)
	: TMeasValue(digits, divide, hysteresis, descrTx, unitTx)
{
	ErrorTx = _errorTx;
	Disable();
}

void TMonitorValue::SetMonitor(uint16_t limitValue, Directions direction, uint16_t timeTicks,
	uint8_t strictPercent, uint16_t strictTicks)
{
	Limit = limitValue;
	Dir = direction;
	MaxCount = (timeTicks < 1) ? 1 : timeTicks;
	if (strictPercent)
	{
		int32_t l = strictPercent;
		l *= (int8_t)Dir;
		l += 100;
		if (l > 0)
		{
			l *= Limit;
			l /= 100;
		}
		else
			l = 0;
		Limit2 = (uint16_t)l;
		MaxCount2 = (strictTicks < 1) ? 1 : strictTicks;
	}
	else
		Limit2 = MaxCount2 = 0;

	Reset();
}

bool TMonitorValue::SetValue(uint16_t value)
{
	TMeasValue::SetValue(value);

	int16_t diff;

	if (MaxCount == 0) return true;	// disabled
	if (MaxCount2 != 0)
	{
		if (Count2 < MaxCount2)
		{
			diff = value - Limit2;
			diff *= (int8_t)Dir;
			if (diff > 0) Count2++;
			else Count2 = 0;
		}
	}

	if (Count < MaxCount)
	{
		diff = value - Limit;
		diff *= (int8_t)Dir;
		if (diff > 0) Count++;
		else Count = 0;
	}

	if (GetCompareResult()) return false;
	return true;
}

void TMonitorValue::PrintLastError1(uint8_t x, uint8_t y) const
{
	lcd_setcursor(x, y);
	x += lcd_string_P(ErrorTx);
	while (x++ < 16) lcd_data(' ');
}

void TMonitorValue::PrintLastError2(uint8_t x, uint8_t y) const
{
	lcd_setcursor(x, y);
	Format(Value, LcdBuff);
	x += lcd_string(LcdBuff);
	switch (Dir)
	{
	case Plus:	lcd_data('>'); break;
	case Minus:	lcd_data('<'); break;
	default: return;
	}
	x++;

	uint16_t lim;
	switch (GetCompareResult())
	{
	case 2: lim = Limit2; break;
	case 1:	lim = Limit; break;
	default: return;
	}
	Format(lim, LcdBuff);
	x += lcd_string(LcdBuff);
	if (UnitTx) x += lcd_string_P(UnitTx);
	while (x++ < 16) lcd_data(' ');
}
