#include <avr/io.h> 

#include "dc.h"
#include "ac.h"
#include "led.h"
#include "eedata.h"
#include "status.h"
#include "adc.h"

enum eTimerActions	{
	actNothing = 0,
	actTimeout = 1,
	actVoltMonitor = 2,
	actVoltAdjust = 4,
};


extern int16_t CurrentDcVoltage;

TStateMachine MachineState;

#define INVALID_TIME 0xffff

TStateMachine::TStateMachine()
{
	State = mstReady;
	AutoStart = false;
}

// Set State Timer
// ticks: Ticks for Timeout or Voltage Adjustment
// what:  Desired Actions, combination of enum timer_actions
// comp_voltage: Compare Voltage or target Voltage for Adjustment
//				> 0 Trigger at Voltage > Compare Voltage, 
//				< 0 Trigger at Voltage < Compare Voltage, 
//				== 0 feature off
// action: Trigger Function
void TStateMachine::SetTimer(uint16_t ticks, uint8_t what, int16_t comp_voltage, FP_StateAction action)
{
	What = what;
	Action = action;
	MaxTime = Time = ticks;
	if (Time == INVALID_TIME) What &= ~(actTimeout | actVoltAdjust);		// Time would be required

	CompVolt = comp_voltage;
	if (CompVolt == 0)
		What &= ~(actVoltMonitor | actVoltAdjust);	// Voltage would be required

	StartVolt = CurrentDcVoltage;
}

eMachineStates TStateMachine::DoTimer(int16_t voltage)
{
	if (mstReady == State && !AutoStart && eeData.auto_on)
	{
		// Autostart
		AutoStart = true;				// nur 1x
		Start();
	}

	uint16_t t = Time;					// use register
	if (t == 0) return State;			// abgelaufen/nicht aktiv

	uint8_t what = actNothing;

	// Timer Funktion
	if (t != INVALID_TIME)
	{
		if (--t == 0) what = What & actTimeout;
	}

	if (What & actVoltMonitor)
	{
		// Voltage Monitor Funktion
		if (CompVolt > 0)
		{
			if (voltage > CompVolt)
				what = actVoltMonitor;		// Ziel erreicht
		}
		else if (CompVolt < 0)
		{
			if (voltage < (-CompVolt))
				what = actVoltMonitor;		// Ziel erreicht
		}
	}

	if (what != actNothing)
	{
		Time = 0;						// Timer deaktivieren
		Action(*this, what);			// Trigger auslsen 
		return State;
	}

	// Voltage Adjust Funktion
	if (What & actVoltAdjust)
	{
		int32_t l = MaxTime - t;
		l *= (CompVolt - StartVolt);
		l /= MaxTime;
		l += StartVolt;
		if (l < 0) l = 0;
		SetDcPwm((uint16_t)l);
	}
	Time = t;

	return State;
}

// private Version
void TStateMachine::StopSoft()
{
	// Reihenfolge:
	// 1. DC Aus
	DisableDc();
	set_run_led(0);

	// 2. Warten bis ZK Spannung gesunken ist
	// 3. AC aus

	// AC nach 5 Sekunden oder wenn ZK Voltage < 20V ist ausschalten
	SetTimer(ADC_SAMPLE_NUM(DC_ADC_MEAN, 5000), actTimeout | actVoltMonitor, -200, &AcOff);
}

void TStateMachine::SetState(eMachineStates new_state)
{
	if (new_state == State) return;
	if (IsError(new_state))
	{
		if (!IsError(State)) StopSoft();
		set_error_led(1);
	}
	else
	{
		set_error_led(0);
	}
	State = new_state;
}

// public Version
void TStateMachine::Start()
{
	if (mstReady == State)
	{
		SetState(mstStarting);
		// TARGET_VOLTAGE_MIN soll in 2 sec. erreicht werden
		SetTimer(ADC_SAMPLE_NUM(DC_ADC_MEAN, 2000), actVoltAdjust | actVoltMonitor, TARGET_VOLTAGE_MIN, &VoltageUp);
		EnableDc();
	}
}

// private Version
void TStateMachine::StopImmediate()
{
	DisableDc();
	disable_ac();
	set_run_led(0);
	if (!IsError(State))
		SetState(mstReady);
}

// public Version
void TStateMachine::Stop()
{
	switch (State)
	{
	case mstStarting:
		StopImmediate();
		break;
	case mstReg:
		StopSoft();
		SetState(mstStopping);
		break;
	default:
		if (IsError(State)) SetState(mstReady);
		break;
	}
}

void TStateMachine::VoltageUp(TStateMachine& s, uint8_t what)
{
	// Zielspannung erreicht -> Normalbetrieb starten
	s.SetState(mstReg);

	// AC mit 2 Sekunden Verzgerung starten damit die Regelung einschwingen kann
	s.SetTimer(ADC_SAMPLE_NUM(DC_ADC_MEAN, 2000), actTimeout, 0, &AcOn);
}

void TStateMachine::AcOn(TStateMachine& s, uint8_t what)
{
	enable_ac();
	set_run_led(1);
}

void TStateMachine::AcOff(TStateMachine& s, uint8_t what)
{
	s.StopImmediate();
}